@@ -1,5 +1,5 @@
 /* pinentry.c - The PIN entry support library
- * Copyright (C) 2002, 2003, 2007, 2008, 2010, 2015, 2016 g10 Code GmbH
+ * Copyright (C) 2002, 2003, 2007, 2008, 2010, 2015, 2016, 2021 g10 Code GmbH
  *
  * This file is part of PINENTRY.
  *
@@ -81,7 +81,7 @@ static const char *flavor_flag;
 /* Because gtk_init removes the --display arg from the command lines
  * and our command line parser is called after gtk_init (so that it
  * does not see gtk specific options) we don't have a way to get hold
- * of the --display option.  Our solution is to remember --disable in
+ * of the --display option.  Our solution is to remember --display in
  * the call to pinentry_have_display and set it then in our
  * parser.  */
 static char *remember_display;
@@ -111,10 +111,15 @@ pinentry_reset (int use_defaults)
   char *default_cf_visi = pinentry.default_cf_visi;
   char *default_tt_visi = pinentry.default_tt_visi;
   char *default_tt_hide = pinentry.default_tt_hide;
+  char *default_capshint = pinentry.default_capshint;
   char *touch_file = pinentry.touch_file;
   unsigned long owner_pid = pinentry.owner_pid;
   int owner_uid = pinentry.owner_uid;
   char *owner_host = pinentry.owner_host;
+  int constraints_enforce = pinentry.constraints_enforce;
+  char *constraints_hint_short = pinentry.constraints_hint_short;
+  char *constraints_hint_long = pinentry.constraints_hint_long;
+  char *constraints_error_title = pinentry.constraints_error_title;
 
   /* These options are set from the command line.  Don't reset
      them.  */
@@ -148,9 +153,13 @@ pinentry_reset (int use_defaults)
       free (pinentry.default_cf_visi);
       free (pinentry.default_tt_visi);
       free (pinentry.default_tt_hide);
+      free (pinentry.default_capshint);
       free (pinentry.touch_file);
       free (pinentry.owner_host);
       free (pinentry.display);
+      free (pinentry.constraints_hint_short);
+      free (pinentry.constraints_hint_long);
+      free (pinentry.constraints_error_title);
     }
 
   free (pinentry.title);
@@ -165,6 +174,7 @@ pinentry_reset (int use_defaults)
   free (pinentry.repeat_error_string);
   free (pinentry.quality_bar);
   free (pinentry.quality_bar_tt);
+  free (pinentry.formatted_passphrase_hint);
   free (pinentry.keyinfo);
   free (pinentry.specific_err_info);
 
@@ -208,10 +218,15 @@ pinentry_reset (int use_defaults)
       pinentry.default_cf_visi = default_cf_visi;
       pinentry.default_tt_visi = default_tt_visi;
       pinentry.default_tt_hide = default_tt_hide;
+      pinentry.default_capshint = default_capshint;
       pinentry.touch_file = touch_file;
       pinentry.owner_pid = owner_pid;
       pinentry.owner_uid = owner_uid;
       pinentry.owner_host = owner_host;
+      pinentry.constraints_enforce = constraints_enforce;
+      pinentry.constraints_hint_short = constraints_hint_short;
+      pinentry.constraints_hint_long = constraints_hint_long;
+      pinentry.constraints_error_title = constraints_error_title;
 
       pinentry.debug = debug;
       pinentry.display = display;
@@ -411,6 +426,32 @@ copy_and_escape (char *buffer, const void *text, size_t textlen)
 }
 
 
+/* Perform percent unescaping in STRING and return the new valid length
+   of the string.  A terminating Nul character is inserted at the end of
+   the unescaped string.
+ */
+static size_t
+do_unescape_inplace (char *s)
+{
+  unsigned char *p, *p0;
+
+  p = p0 = s;
+  while (*s)
+    {
+      if (*s == '%' && s[1] && s[2])
+        {
+          s++;
+          *p++ = xtoi_2 (s);
+          s += 2;
+        }
+      else
+        *p++ = *s++;
+    }
+  *p = 0;
+
+  return (p - p0);
+}
+
 
 /* Return a malloced copy of the commandline for PID.  If this is not
  * possible NULL is returned.  */
@@ -626,6 +667,71 @@ pinentry_inq_quality (pinentry_t pin, const char *passphrase, size_t length)
   return value;
 }
 
+
+/* Run a checkpin inquiry */
+char *
+pinentry_inq_checkpin (pinentry_t pin, const char *passphrase, size_t length)
+{
+  assuan_context_t ctx = pin->ctx_assuan;
+  const char prefix[] = "INQUIRE CHECKPIN ";
+  char *command;
+  char *line;
+  size_t linelen;
+  int gotvalue = 0;
+  char *value = NULL;
+  int rc;
+
+  if (!ctx)
+    return 0; /* Can't run the callback.  */
+
+  if (length > 300)
+    length = 300;  /* Limit so that it definitely fits into an Assuan
+                      line.  */
+
+  command = secmem_malloc (strlen (prefix) + 3*length + 1);
+  if (!command)
+    return 0;
+  strcpy (command, prefix);
+  copy_and_escape (command + strlen(command), passphrase, length);
+  rc = assuan_write_line (ctx, command);
+  secmem_free (command);
+  if (rc)
+    {
+      fprintf (stderr, "ASSUAN WRITE LINE failed: rc=%d\n", rc);
+      return 0;
+    }
+
+  for (;;)
+    {
+      do
+        {
+          rc = assuan_read_line (ctx, &line, &linelen);
+          if (rc)
+            {
+              fprintf (stderr, "ASSUAN READ LINE failed: rc=%d\n", rc);
+              return 0;
+            }
+        }
+      while (*line == '#' || !linelen);
+      if (line[0] == 'E' && line[1] == 'N' && line[2] == 'D'
+          && (!line[3] || line[3] == ' '))
+        break; /* END command received*/
+      if (line[0] == 'C' && line[1] == 'A' && line[2] == 'N'
+          && (!line[3] || line[3] == ' '))
+        break; /* CAN command received*/
+      if (line[0] == 'E' && line[1] == 'R' && line[2] == 'R'
+          && (!line[3] || line[3] == ' '))
+        break; /* ERR command received*/
+      if (line[0] != 'D' || line[1] != ' ' || linelen < 3 || gotvalue)
+        continue;
+      gotvalue = 1;
+      value = strdup (line + 2);
+    }
+
+  return value;
+}
+
+
 /* Run a genpin inquiry */
 char *
 pinentry_inq_genpin (pinentry_t pin)
@@ -656,6 +762,7 @@ pinentry_inq_genpin (pinentry_t pin)
           if (rc)
             {
               fprintf (stderr, "ASSUAN READ LINE failed: rc=%d\n", rc);
+              free (value);
               return 0;
             }
         }
@@ -1247,6 +1354,12 @@ option_handler (assuan_context_t ctx, const char *key, const char *value)
       if (!pinentry.default_tt_hide)
 	return gpg_error_from_syserror ();
     }
+  else if (!strcmp (key, "default-capshint"))
+    {
+      pinentry.default_capshint = strdup (value);
+      if (!pinentry.default_capshint)
+	return gpg_error_from_syserror ();
+    }
   else if (!strcmp (key, "allow-external-password-cache") && !*value)
     {
       pinentry.allow_external_password_cache = 1;
@@ -1266,6 +1379,48 @@ option_handler (assuan_context_t ctx, const char *key, const char *value)
       if (!pinentry.invisible_char)
 	return gpg_error_from_syserror ();
     }
+  else if (!strcmp (key, "formatted-passphrase") && !*value)
+    {
+      pinentry.formatted_passphrase = 1;
+    }
+  else if (!strcmp (key, "formatted-passphrase-hint"))
+    {
+      if (pinentry.formatted_passphrase_hint)
+        free (pinentry.formatted_passphrase_hint);
+      pinentry.formatted_passphrase_hint = strdup (value);
+      if (!pinentry.formatted_passphrase_hint)
+	return gpg_error_from_syserror ();
+      do_unescape_inplace(pinentry.formatted_passphrase_hint);
+    }
+  else if (!strcmp (key, "constraints-enforce") && !*value)
+    pinentry.constraints_enforce = 1;
+  else if (!strcmp (key, "constraints-hint-short"))
+    {
+      if (pinentry.constraints_hint_short)
+        free (pinentry.constraints_hint_short);
+      pinentry.constraints_hint_short = strdup (value);
+      if (!pinentry.constraints_hint_short)
+	return gpg_error_from_syserror ();
+      do_unescape_inplace(pinentry.constraints_hint_short);
+    }
+  else if (!strcmp (key, "constraints-hint-long"))
+    {
+      if (pinentry.constraints_hint_long)
+        free (pinentry.constraints_hint_long);
+      pinentry.constraints_hint_long = strdup (value);
+      if (!pinentry.constraints_hint_long)
+	return gpg_error_from_syserror ();
+      do_unescape_inplace(pinentry.constraints_hint_long);
+    }
+  else if (!strcmp (key, "constraints-error-title"))
+    {
+      if (pinentry.constraints_error_title)
+        free (pinentry.constraints_error_title);
+      pinentry.constraints_error_title = strdup (value);
+      if (!pinentry.constraints_error_title)
+	return gpg_error_from_syserror ();
+      do_unescape_inplace(pinentry.constraints_error_title);
+    }
   else
     return gpg_error (GPG_ERR_UNKNOWN_OPTION);
   return 0;
